/*****************************************************************************
Spartan-3 Starter Board to Parallel Port Interface
  ADT representing a string of binary digits.

  Written by Eirik Bakke, January 14, 2006
  See http://www.mit.edu/~ebakke/parint for documentation
  
******************************************************************************/


#include <assert.h>
#include <stddef.h>
#include <malloc.h>
#include "binval.h"
#include "stdio.h"

struct Binval {
	int width;
	int *bits;
	char *tostr;
	int issubstring;
	Binval_T next_substring;
};

enum {
	TRUE = 1,
	FALSE = 0,
	BUF_SIZE = 4096
};

static Binval_T Binval_new_init(int width, int issubstring) {
	Binval_T ret;
	assert(width > 0);
	ret = (Binval_T) malloc(sizeof(struct Binval));
	assert(ret != NULL);
	ret->width = width;
	ret->issubstring = issubstring;
	ret->tostr = (char *) calloc(width + 1, sizeof(char));
	assert(ret->tostr != NULL);
	ret->next_substring = NULL;
	if (issubstring) {
		ret->bits = NULL;
	} else {
		ret->bits  = (int *) calloc(width, sizeof(int));
		assert(ret->bits != NULL);
	}
	return ret;
}

Binval_T Binval_new(int width) {
	return Binval_new_init(width, FALSE);
}

Binval_T Binval_newval(int width, int val) {
	Binval_T ret;
	ret = Binval_new(width);
	Binval_set(ret, val);
	return ret;
}

Binval_T Binval_newsubstring(Binval_T b, int pos, int width) {
	Binval_T ret;
	assert(b != NULL);
	assert(width > 0);
	assert(pos >= 0 && pos + width - 1 < b->width);
	ret = Binval_new_init(width, TRUE);
	ret->bits = &(b->bits[pos]);
	ret->next_substring = b->next_substring;
	b->next_substring = ret;
	return ret;
}

static void Binval_freenext(Binval_T b) {
	if (!b->issubstring) {
		free(b->bits);
	}
	if (b->next_substring != NULL)
		Binval_freenext(b->next_substring);
	free(b->tostr);
	free(b);
}

void Binval_free(Binval_T b) {
	assert(b != NULL);
	assert(!b->issubstring);
	Binval_freenext(b);
}

void Binval_set(Binval_T b, int val) {
	int i;
	assert(b != NULL);
	for (i = 0; i < b->width; i++) {
		b->bits[i] = (val & 1) != 0;
		val >>= 1;
	}
}

int Binval_get(Binval_T b) {
	int i, s = 0;
	for (i = b->width - 1; i >= 0; i--) {
		s <<= 1;
		s |= (b->bits[i] != 0);
	}
	return s;
}

int Binval_equal(Binval_T a, Binval_T b) {
	int i;
	Binval_T longer, shorter;
	assert(a != NULL && b != NULL);

	longer  = (a->width > b->width) ? a : b;
	shorter = (a->width > b->width) ? b : a;

	for (i = 0; i < longer->width; i++) {
		if (
			(i >= shorter->width) ?
				(longer->bits[i] != 0) :
				((longer->bits[i] == 0) != (shorter->bits[i] == 0))
		)
			return FALSE;
	}
	return TRUE;
}

char *Binval_tostring(Binval_T b) {
	int i;
	assert(b != NULL);
	for (i = 0; i < b->width; i++)
		(b->tostr)[b->width - 1 - i] = (b->bits[i] == 0) ? '0' : '1';
	return b->tostr;
}

int *Binval_toarray(Binval_T b) {
	assert(b != NULL);
	return b->bits;
}

int *Binval_digit(Binval_T b, int i) {
	assert(b != NULL);
	assert(i >= 0 && i < b->width);
	return &(b->bits[i]);
}

/*
int main(int argc, char **argv) {
	Binval_T a, b, c, d;
	char *s;
	b = Binval_newval(12, 140);
	assert(strcmp(Binval_tostring(b), "000010001100") == 0);
	a = Binval_newval(14, 140);
	c = Binval_newsubstring(a, 3, 2);
	d = Binval_newsubstring(c, 0, 1);
	assert(Binval_equal(a, b));
	Binval_set(a, 0);
	assert(!Binval_equal(a, b));
	Binval_set(c, 2);
	Binval_set(b, 16);
	assert(strcmp(Binval_tostring(a), "00000000010000") == 0);
	assert(strcmp(Binval_tostring(b),   "000000010000") == 0);
	assert(strcmp(Binval_tostring(c),             "10") == 0);
	assert(Binval_equal(a, b));
	Binval_set(c, 3 + 8);
	Binval_set(b, 16 + 8);
	assert(strcmp(Binval_tostring(a), "00000000011000") == 0);
	assert(strcmp(Binval_tostring(b),   "000000011000") == 0);
	assert(strcmp(Binval_tostring(c),             "11") == 0);
	assert(Binval_equal(a, b));
	Binval_set(d, 0);
	assert(!Binval_equal(a, b));
	Binval_free(a);
	Binval_free(b);
	printf("Success.\n");
	while(1);
	return 0;
}
*/
