/*****************************************************************************
Spartan-3 Starter Board to Parallel Port Interface
  Interactive command-line tool for setting and getting FPGA signals.
  
  Expects a file containing a list of signal directions, widths and names to be
  specified on the command line. With no additional arguments, takes input from
  stdin where named signals can be set, triggered, asserted, or queried. With an
  additional file name specified, generates a Verilog stub for the same given
  list of signals.

  Written by Eirik Bakke, January 14, 2006
  See http://www.mit.edu/~ebakke/parint for documentation
  
******************************************************************************/

#include <assert.h>
#include <stddef.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "parint.h"
#include "binval.h"

enum {
	FALSE = 0,
	TRUE  = 1,
	BUF_SIZE = 65536,
	COMMENT_CHAR = '#',
	UPDATE_MANUAL = 0,
	UPDATE_AUTO,
	UPDATE_INTERACTIVE,
	UPDATE_LAST,
	KEYWORD_QUIT = 0, 
	KEYWORD_UPDATE,
	KEYWORD_UPDATE_LEVEL,
	KEYWORD_LAST
};

static const char *keywords[3] = {"quit", "update", "update_level"};
static char dummy[2];

typedef struct {
	char isinput; /* From the hardware's point of view. */
	char *name;
	int  width;
	Binval_T val;
} sigentry;

static int nosigs = 0;
static sigentry sigs[BUF_SIZE];
static int w_in = -1, w_out = -1;
static int uplev = UPDATE_INTERACTIVE;

static void finish(int success) {
	int i;
    Parint_shutdown();
	for (i = 0; i < nosigs; i++) {
		free(sigs[i].name);
		sigs[i].name = NULL;
	}
	nosigs = 0;
    exit(success ? EXIT_SUCCESS : EXIT_FAILURE);
}

static void error(int fatal, const char *format, ...) {
    va_list ap;
	fprintf(stderr, (fatal) ? "Error: " : "Warning: ");
    va_start(ap, format);
    vfprintf(stderr, format, ap);
    va_end(ap);
    fprintf(stderr, ".\n");
	if (fatal)
		finish(FALSE);
}

static sigentry *lookupsig(char *name) {
	int i;
	for (i = 0; i < nosigs; i++)
		if (strcmp(name, sigs[i].name) == 0)
			return &(sigs[i]);
	return NULL;
}

static void printsignals(void) {
	int i;
	for (i = 0; i < nosigs; i++) {
		printf("%c\t%s\t= %s", (sigs[i].isinput) ? 'i' : 'o',
				sigs[i].name, Binval_tostring(sigs[i].val));
		if (sigs[i].width > 1)
			printf(" (%o)", Binval_get(sigs[i].val));
		printf("\n");
	}
}

static void rtrim(char *str) {
	size_t i, j, len = strlen(str);
	for (i = 0, j = len - 1; i < len && isspace(str[j]); i++, j--) {
		str[j] = '\0';
	}
}

static void interact(void) {
	static char ibuf[BUF_SIZE], kbuf[BUF_SIZE];
	int inval;
	sigentry *ent;
	char c = 0;
	int report, interactive;
	
	while (TRUE) {
		report      = FALSE;
		interactive = (uplev >= UPDATE_INTERACTIVE);

		fgets(ibuf, sizeof(ibuf), stdin);
		rtrim(ibuf);

		if (feof(stdin) || strcmp(ibuf, keywords[KEYWORD_QUIT]) == 0) {
			break;
		} else if (strcmp(ibuf, keywords[KEYWORD_UPDATE]) == 0) {
			Parint_update();
		} else if (ibuf[0] == COMMENT_CHAR) {
			/* Comment; do nothing. */
		} else if (sscanf_s(ibuf, "%s %o %1s", kbuf, sizeof(kbuf), &inval, dummy, 2) == 2) {
			if (strcmp(kbuf, keywords[KEYWORD_UPDATE_LEVEL]) == 0) {
				if (inval >= 0 && inval < UPDATE_LAST)
					uplev = inval;
				else
					error(TRUE, "Invalid update level %d", uplev);
			} else {
				/* Signal name plus value */
				if ((ent = lookupsig(kbuf)) == NULL) {
					error(!interactive, "Cannot set unknown signal \"%s\"", kbuf);
				} else {
					if (ent->isinput) {
						Binval_set(ent->val, inval);
						if (uplev >= UPDATE_AUTO)
							Parint_update();
						report = TRUE;
					} else {
						if (Binval_get(ent->val) != inval)
							error(!interactive, "Signal assertion \"%s == %o\" failed", kbuf, inval);
					}
				}
			}
		} else if (sscanf_s(ibuf, "%s %1s", kbuf, sizeof(kbuf), dummy, 2) == 1) {
			/* Signal name alone */
			if ((ent = lookupsig(kbuf)) == NULL) {
				error(!interactive, "Cannot trigger unknown signal \"%s\"", kbuf);
			} else {
				if (ent->isinput) {
					/* Trigger signal */
					Binval_set(ent->val, ~Binval_get(ent->val));
					Parint_update();
					/*Parint_sleep(1);*/
					Binval_set(ent->val, ~Binval_get(ent->val));
					Parint_update();
					report = TRUE;
				} else {
					/* Print output value of signal */
					if (uplev >= UPDATE_AUTO)
						Parint_update();
					printf("%o\n", Binval_get(ent->val));
				}
			}
		} else if (sscanf_s(ibuf, "%1s", dummy, 2) > 0) {
			error(!interactive, "Invalid syntax in command \"%s\"", ibuf);
		} else {
			assert(strlen(ibuf) == 0);
			report = TRUE;
		}
		if (interactive) {
			if (report) {
				Parint_update();
				printsignals();
				printf("\n");
			}
		}
	}
}

static void readSigList(FILE *f) {
	static char lbuf[BUF_SIZE], nbuf[BUF_SIZE];
	size_t tmp = 0;
	sigentry *ent;
	int i;
	char io;
	
	w_in = w_out = 0;
	nosigs = 0;

	while (TRUE) {
		if (fgets(lbuf, sizeof(lbuf), f) == NULL)
			break;
		rtrim(lbuf);
		if (lbuf[0] == COMMENT_CHAR || lbuf[0] == '\0')
			continue;

		assert(nosigs < sizeof(sigs) / sizeof(sigentry));
		ent = &(sigs[nosigs]);

		if (sscanf_s(lbuf, "%c %d %s %1s", &io, 1, &(ent->width), nbuf, sizeof(nbuf), dummy, 2) != 3)
			error(TRUE, "Malformed input line \"%s\"", lbuf);
		if (io != 'i' && io != 'o')
			error(TRUE, "First column must be either 'i' or 'o'");
		if (lookupsig(nbuf) != NULL)
			error(TRUE, "Duplicate definition of signal \"%s\"", nbuf);
		for (i = 0; i < KEYWORD_LAST; i++)
			if (strcmp(nbuf, keywords[i]) == 0)
				error(TRUE, "Invalid signal name: \"%s\" is a reserved word", nbuf);

		ent->isinput = (io == 'i');
		tmp = strlen(nbuf) + 1;
		ent->name = (char *) malloc(tmp);
		assert(ent->name != NULL);
		strcpy_s(ent->name, tmp, nbuf);
		ent->val = NULL;

		*((ent->isinput) ? (&w_in) : (&w_out)) += ent->width;

		nosigs++;
	}
}

static writeSiglist(FILE *f, int dir_in) {
	sigentry *i;
	int j, first = TRUE;

	for (i = sigs, j = (dir_in) ? w_in : w_out; i < sigs + nosigs; i++) {
		if (i->isinput == dir_in) {
			j -= i->width;
			fprintf(f, (first) ? "%s" : ", %s", i->name);
			first = FALSE;
		}
	}
}

static void writeVerilog(FILE *f) {
	assert(f != NULL);
	fprintf(f, "\t/******* Auto-generated, do not modify *******/\n");
	fprintf(f, "\tParint #(%d, %d) mparint(tdi, tck, tms, tdo,\n\t\t{", w_in, w_out);
	writeSiglist(f, TRUE);
	fprintf(f, "},\n\t\t{");
	writeSiglist(f, FALSE);
	fprintf(f, "});\n");
	fprintf(f, "\t/*********************************************/\n");
}

static int initInterface(void) {
	int bin = w_in, bout = w_out;
	sigentry *i;

	if (Parint_init(w_in, w_out))
		return -1;

	for (i = sigs; i < sigs + nosigs; i++) {
		if (i->isinput) {
			bin -= i->width;
			i->val = Binval_newsubstring(Parint_inputs(), bin, i->width);
		} else {
			bout -= i->width;
			i->val = Binval_newsubstring(Parint_outputs(), bout, i->width);
		}
	}
	return 0;
}

static void usage(char *argv0) {
	fprintf(stderr, 
		"Spartan S3 Starter Kit to Parallel Port Interface\n"
		"  by Eirik Bakke (ebakke@mit.edu)\n"
		"  http://www.mit.edu/~ebakke/parint\n\n"
		"Usage: %s [<signals list input file> [<verilog output file>] | uninstall]\n\n"
		"Columns in signals list:\n  i/o? size name\n\n",
		argv0
		);
	finish(FALSE);
}

int main(int argc, char **argv) {
	FILE *sigfile = NULL;
	FILE *vfile   = NULL;

	if (argc == 2 && strcmp(argv[1], "uninstall") == 0) {
		Parint_uninstall();
		finish(TRUE);
	}

	if (argc != 2 && argc != 3)
		usage(argv[0]);

	if (fopen_s(&sigfile, argv[1], "r") != 0)
		error(TRUE, "Could not open signals list input file %s", argv[1]);

	if (argc >= 3 && fopen_s(&vfile, argv[2], "w") != 0)
		error(TRUE, "Could not open Verilog output file %s", argv[2]);

	readSigList(sigfile);
	fclose(sigfile);

    if (initInterface())
		error(TRUE, "Failed to initialize Parint");

	if (vfile != NULL) {
		writeVerilog(vfile);
		fclose(vfile);
		finish(TRUE);
	}

	interact();
	finish(TRUE);
}
